#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdbool.h>
#include <avr/eeprom.h>
#include <stdlib.h>


#define F_CPU 4000000L

// EEMEM wird bei aktuellen Versionen der avr-lib in eeprom.h definiert
// hier: definiere falls noch nicht bekannt ("alte" avr-libc)
#ifndef EEMEM
// alle Textstellen EEMEM im Quellcode durch __attribute__ ... ersetzen
#define EEMEM  __attribute__ ((section (".eeprom")))
#endif

//Ports

#define ENA_PORT 	PORTB
#define ENA_DDR		DDRB
#define ENA_1 		2
#define ENA_2		3
#define ENA_3		4
#define ENA_4		5

uint8_t EnablePins[4]	= {ENA_4,ENA_3,ENA_2,ENA_1};//Reihenfolge der EnablePins zuweisen 

#define SEG_G		6
#define SEG_F		5
#define SEG_E		4
#define SEG_D		3
#define SEG_C		2
#define SEG_B		1
#define SEG_A		0

#define SEG_PORT	PORTD
#define SEG_DDR		DDRD
#define SEG_ALL		0x7F

/*
#define DOT_PORT	PORTB
#define DOT_DDR		DDRB
#define DOT_PIN		0x01
*/

#define RELAIS_PORT	PORTB
#define RELAIS_DDR	DDRB
#define RELAIS_PIN	1

#define PIEZO_PORT	PORTB
#define PIEZO_DDR	DDRB
#define PIEZO_PIN	0


#define TASTER_DDR	DDRC
#define TASTER_PORT	PORTC
#define TASTER_PIN	PINC

#define TASTER_UP		4
#define TASTER_DOWN		5
#define TASTER_START	3	

//Makros
#define Relais_Aus()	RELAIS_PORT |= (1<<RELAIS_PIN)
#define Relais_Ein()	RELAIS_PORT	&= ~(1<<RELAIS_PIN)


//ProgrammPositionen / Zustnde
#define ST_IDLE		0
#define ST_RUN		1
#define ST_SETUP	2
#define ST_STOP		3
#define ST_PAUSE	4


//Variable global

unsigned long int i,j;

uint8_t Ziffern_Feld[10] = 	{
							0x40,//0
							0xF9,//1
							0x24,//2
							0x30,//3
							0x19,//4
							0x12,//5
							0x02,//6
							0x78,//7
							0x00,//8
							0x10,//9
							};

uint8_t Anzeige_Pos = 0;
uint8_t Zahl = 0;
uint8_t State = 0;
uint16_t Zeit = 0;
uint16_t EEPromZeit[10] EEMEM = {100,100,100,100,100,100,100,100,100,100,100};
uint8_t LastTime = 0;

//Prototypen
uint8_t Write_Ziffer_Segment(uint8_t Ziffer,uint8_t Pos);
uint16_t Bin2Bcd(uint16_t BinZahl);
uint16_t Bin2BcdClk(uint16_t BinZahl);
bool ShowDisplay = true;


//***************************************************************************************************
//*****************************  Interrupts *********************************************************
//***************************************************************************************************
ISR(TIMER1_CAPT_vect)
{
	static uint16_t Debounce; //Zyklen fr TastenEntprellung
	static uint16_t Refresh_Delay; //Zyklen fr Aktualisierung Anzeige
	static uint16_t Time_Delay; //Zyklen fr Zeitzhlung
	static uint16_t Setup_Refresh;

	static bool		Up, Up_Prev;
	static bool		Down, Down_Prev;
	static bool		Double, Double_Prev;
	static bool		Start, Start_Prev;

	static uint8_t Ziffer;
	static char Zeichen;
	static uint8_t Anzeige[4];
	static uint16_t Temp;
	static uint16_t UpKey_HoldTime;
	static uint16_t DownKey_HoldTime;
	
	Debounce++;
	
	if (Debounce == 100)
	{
		
		Debounce = 0;

		Up = false;
		Down = false;
		Double = false;
		Start = false;


		if ((TASTER_PIN & (1<<TASTER_UP) ) == 0)
		{
			Up = true;
		}
		
		if ((TASTER_PIN & (1<<TASTER_DOWN) ) == 0)
		{
			Down = true;
		}

		if ((TASTER_PIN & (1<<TASTER_START) ) == 0)
		{
			Start = true;
		}

		if ( (Up == true) && (Down == true) )
		{
			Double = true;
		}

		if ( (Up == true) && (Up_Prev == true) )
		{
			if (UpKey_HoldTime < 100)
				UpKey_HoldTime++;
		}
		else 
		{
			UpKey_HoldTime = 0;
		}

		if ( (Down == true) && (Down_Prev == true) )
		{
			if (DownKey_HoldTime < 100)
				DownKey_HoldTime++;
		}
		else 
		{
			DownKey_HoldTime = 0;
		}

		switch (State)
		{

			case ST_IDLE 	: 	/*
								if ( (Up == true) && (Up_Prev == false) && (Down == false) && (Zeit > 0) )
								{
									State = ST_RUN;
									Relais_Ein();
								}*/

								if ( (Start == true) && (Start_Prev == false) && (Up == false) && (Down == false) && (Zeit > 0) )
								{
									State = ST_RUN;
									Relais_Ein();
								}

								if ( (Double == true) && (Double_Prev == false) )//steigende Flanke beide Tasten gedrckt
								{
									State = ST_SETUP;
								}

								break;

			case ST_STOP 	: 	if ( ( (Start == true) && (Start_Prev == false) && (Up == false) && (Down == false) )
									||
									 ( (Start == false) && (Up == true) && (Up_Prev == false) && (Down == false) )
									||
									 ( (Start == false) && (Up == false) && (Down == true) && (Down_Prev == false) )
								   )
								{
									State = ST_IDLE;
									Zeit = eeprom_read_byte(&EEPromZeit[0]);
								}
								break;

			case ST_PAUSE 	: 	if ( (Start == true) && (Start_Prev == false) && (Up == false) && (Down == false) && (Zeit > 0) )
								{
									State = ST_RUN;
									Relais_Ein();
								}
								
								if ( ( ((Down == true) && (Down_Prev == false) && (Up == false))
									|| ((Up == true) && (Up_Prev == false) && (Down == false)) )
									&& (Zeit > 0) )
								{
									State = ST_IDLE;
									Zeit = eeprom_read_byte(&EEPromZeit[0]);
								}

								break;

			case ST_RUN		: 	if ( (Start == true) && (Start_Prev == false) && (Up == false) && (Down == false) && (Zeit > 0) )
								{
									State = ST_PAUSE;
									Relais_Aus();
								}
								break;

			case ST_SETUP	: 	ShowDisplay = !ShowDisplay; //Blinken
								
								if ( (Double == true) && (Double_Prev == false) )//steigende Flanke beide Tasten gedrckt
								{
									State = ST_IDLE;

									ShowDisplay = true; //Display auf jeden Fall wieder an --> nicht mehr blinken
									eeprom_write_byte(&EEPromZeit, Zeit);
								}

								if ( (Up == true) && (Down == false) )
								{
									if ( (Zeit<5999) && (UpKey_HoldTime < 15) ) 
										Zeit++;
									else 
									{
										if ( (Zeit < 5995) && (UpKey_HoldTime >= 15) )
										{
											Zeit += 5;
										}
									}

								}
		
								if ( (Up == false) && (Down == true) )
								{
									if ( (Zeit>0) && (DownKey_HoldTime < 15) ) 
										Zeit--;
									else
									{
										if ( (Zeit > 5) && (DownKey_HoldTime  >= 15) )
										{
											Zeit -= 5;
										}
									}
								}
								
								break;
		}


		Up_Prev = Up;
		Down_Prev = Down;
		Double_Prev = Double;
		Start_Prev = Start;

			
	}//#if (Debounce == 100)

	
	Refresh_Delay++;

	if (Refresh_Delay == 5)//alle 5ms Aktualisierung der Anzeige
	{
		Refresh_Delay = 0;

		if (Anzeige_Pos == 4) 
		{
			Anzeige_Pos = 0;
			Temp = Bin2BcdClk(Zeit);
			Anzeige[0] = (uint8_t)Temp & 0x0F;
			Anzeige[1] = (uint8_t)(Temp >> 4) & 0x0F;
			Anzeige[2] = (uint8_t)(Temp >> 8) & 0x0F;
			Anzeige[3] = (uint8_t)(Temp >> 12) & 0x0F;
		}

		Write_Ziffer_Segment(Anzeige[Anzeige_Pos], Anzeige_Pos);

		Anzeige_Pos++;

	}//#if (Refresh_Delay == 5)

	if (State == ST_RUN)
	{

		Time_Delay++;

		if (Time_Delay == 1000)
		{
			Time_Delay = 0;

			if (Zeit > 0) Zeit--;

			if (Zeit == 0) 
			{
				State = ST_STOP;
				Relais_Aus();//RELAIS_PORT &= ~(1<<RELAIS_PIN); //Relais ausschalten
			}

		}//#if (Time_Delay == 1000)
		

	}//#if (State == ST_RUN)

}


uint8_t Write_Ziffer_Segment(uint8_t Ziffer,uint8_t Pos)
{
	if (ShowDisplay)
	{
		if ( (Pos >= 0) && (Pos <= 3) ) 
		{
			if ( (Ziffer >=0) && (Ziffer <= 9) )
			{
				ENA_PORT |= ( (1<<ENA_1) | (1<<ENA_2) | (1<<ENA_3) | (1<<ENA_4) ); //jeglichen Enable zurcknehmen (invers)

				SEG_PORT = Ziffern_Feld[Ziffer]; //Zeichen auf Bus legen

				//ENA_PORT &= ~( (1<<Pos)<<2 );//Anzeigeelement aktivieren
				ENA_PORT &= ~(1<<EnablePins[Pos]);

			return 0;
			}
			else //Fehler keine Ziffer
			{
				SEG_PORT = 0x4F;
				ENA_PORT &= ~(1<<EnablePins[Pos]);
			}

		
		}
	}
	else
	{
		ENA_PORT |= ( (1<<ENA_1) | (1<<ENA_2) | (1<<ENA_3) | (1<<ENA_4) ); //jeglichen Enable zurcknehmen (invers)
	}

return 1;
}

uint16_t Bin2Bcd(uint16_t BinZahl) //Umwandlung Uint16 in BCD gepackt 
{									//Rckgabe in 16Bit Uint im Format
									//Z0Z0Z0Z0 Z1Z1Z1Z1 Z2Z2Z2Z2 Z3Z3Z3Z3
									//angepasste Version 0 bis 999
	div_t Ergebnis;
	uint8_t BCD[4];

	Ergebnis = div(BinZahl,10000);
	Ergebnis = div(Ergebnis.rem,1000);
	BCD[3]	 = Ergebnis.quot;
	Ergebnis = div(Ergebnis.rem,100);
	BCD[2] = Ergebnis.quot;
	Ergebnis = div(Ergebnis.rem,10);
	BCD[1] = Ergebnis.quot;
	BCD[0] = Ergebnis.rem;

	return (uint16_t) (BCD[3]<<12)|(BCD[2]<<8)|(BCD[1]<<4)|BCD[0];
}

uint16_t Bin2BcdClk(uint16_t BinZahl) //Umwandlung Uint16 in BCD gepackt Uhrzeitformat min:min:sek:sek
{									//Rckgabe in 16Bit Uint im Format
									//Z0Z0Z0Z0 Z1Z1Z1Z1 Z2Z2Z2Z2 Z3Z3Z3Z3
									//angepasste Version 0 bis 9999
	div_t Ergebnis;
	uint8_t BCD[4];

	Ergebnis = div(BinZahl,10000);
	Ergebnis = div(Ergebnis.rem,6000);
	Ergebnis = div(Ergebnis.rem,600);
	BCD[3]	 = Ergebnis.quot;
	Ergebnis = div(Ergebnis.rem,60);
	BCD[2] = Ergebnis.quot;
	Ergebnis = div(Ergebnis.rem,10);
	BCD[1] = Ergebnis.quot;
	BCD[0] = Ergebnis.rem;

	return (uint16_t) (BCD[3]<<12)|(BCD[2]<<8)|(BCD[1]<<4)|BCD[0];
}


void init()
{
	ENA_DDR |= (1<<ENA_1) | (1<<ENA_2) | (1<<ENA_3) | (1<<ENA_4); //Set enable Pin 7Segment as Output

	ENA_PORT |= ( (1<<ENA_1) | (1<<ENA_2) | (1<<ENA_3) | (1<<ENA_4) ); //Alle Enalble High (invers)

	SEG_DDR |= SEG_ALL; //Pins fr 7Segment als Ausgang

	SEG_PORT &= ~SEG_ALL; //Alle 7SegmentAusgnge Low

	RELAIS_DDR |= (1<<RELAIS_PIN); //RelaisPin als Ausgang

	Relais_Aus();

	TASTER_DDR &= ~( (1<<TASTER_UP) | (1<<TASTER_DOWN) | (1<<TASTER_START)  ); //Taster Pins als Eingang

	TASTER_PORT |= ( (1<<TASTER_UP) | (1<<TASTER_DOWN) | (1<<TASTER_START) ); //PullUps an TasterPins aktivieren

	Zeit = eeprom_read_byte(&EEPromZeit[0]);
}

void Timer_Init()
{
	/*
	// Timer 0 CTC-Mode 1ms bei 8MHz
	TCNT0 = 0x00;
	OCR0A = 124;
	OCR0B = 0x00;
	TCCR0A = 0x02; //CTC on CompareMatchA
	TCCR0B = 0x03; //Prescaler 64 
	TIMSK |= (1<<OCIE0A);//Timer0 Interrupt on Compare MatchA
	*/

	
	//Timer Counter 1 CTC, Top = ICR, 1ms @ 4MHz
	TCNT1H	= 	0x00;
	TCNT1L	= 	0x00;

	OCR1AH  =   0x00;
	OCR1AL  =   0x00;
	OCR1BH  =   0x00;
	OCR1BL  =   0x00;

	ICR1H	= 	0x0F;
	ICR1L	= 	0x9F;	//ICR = 3999

	TCCR1A 	= 	0x00;	//CTC Clear on ICR
	TCCR1B 	= 	0x19;	//CTC Clear on ICR, Prescaler 1

	TIMSK  	=	0x20;	//enable ICR-Int

}



void main()
{

	init();
	Timer_Init();
	State = ST_IDLE;
	asm("sei");

	while (1)
	{


	}
}
